#ifndef _COMMANDS_
#define _COMMANDS_

#ifndef _TYPES_
#include "types.h"
#endif

#include "macros.h"

/********************************************************************************
 *																				*
 *		command types															*
 *																				*
 ********************************************************************************/

typedef struct COMMAND
{
	CHAR		*name;
	VOID		(*parse_function)();
	VOID		(*exec_function)();
	
} COMMAND;

typedef struct COMMAND_DEFINITION
{
	WORD	next;
	COMMAND	*definition;

} COMMAND_DEFINITION;

typedef struct VCOMMAND_TABLE
{
	WORD	size;
	BOOL	running;
	
	COMMAND_DEFINITION commands[VariableArraySize];
	
} VCOMMAND_TABLE;

typedef struct COMMAND_LINE
{
	WORD	|channel;					/* to wait until command received */
	WORD	size;						/* size of command */
	CHAR	line[StringLength];			/* characters of command */
	
} COMMAND_LINE;

typedef struct VCOMMAND_PACKET
{
	COMMAND_LINE	*command_dest;
	COMMAND_LINE	*command_source;	
	
	VCLEAN_STRING	command;

} VCOMMAND_PACKET;

/********************************************************************************
 *																				*
 *		exported parse and exec routines										*
 *																				*
 ********************************************************************************/

extern VOID parse_e		(CHAR **string_handle, COMMAND *command);
extern VOID parse_n		(CHAR **string_handle, COMMAND *command);
extern VOID parse_s		(CHAR **string_handle, COMMAND *command);
extern VOID parse_ident	(CHAR **string_handle, COMMAND *command);
extern VOID parse_word	(CHAR **string_handle, COMMAND *command);
extern VOID parse_nn	(CHAR **string_handle, COMMAND *command);

/********************************************************************************
 *																				*
 *		exported error messages													*
 *																				*
 ********************************************************************************/

extern VOID missing_string	(VOID);
extern VOID missing_ident	(VOID);
extern VOID missing_number	(VOID);
extern VOID missing_address	(VOID);

/********************************************************************************
 *																				*
 *		exported command functions												*
 *																				*
 ********************************************************************************/

extern VCOMMAND_TABLE	*get_command_table		(WORD size);
extern BOOL				define_commands			(VCOMMAND_TABLE *command_table, COMMAND *commands, WORD amount);
extern BOOL				exec_command			(VCOMMAND_TABLE *command_table, const CHAR *command_line);

extern VOID				put_command				(WORD dest, WORD command, CHAR *message);
extern VOID				execute_commands		(VCOMMAND_TABLE *command_table, CHAR *prompt);

#endif
