#ifndef _GRAPHS_
#define _GRAPHS_

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

#ifndef _PACKETS_
#include "packets.h"
#endif

/********************************************************************************
 *																				*
 *		graph packets															*
 *																				*
 ********************************************************************************/

/*	For id's 0 acts as NULL: this means that no id is available, lists of id's are
	terminated by 0
*/

typedef struct GRAPH_ID
{
	WORD	dest;
	WORD	source;
	
} GRAPH_ID;

typedef struct REQUEST_PACKET
{
	WORD		id;
	REQUEST		request;
		
} REQUEST_PACKET;

typedef struct DECREF_PACKET
{
	GRAPH_ID	id;
	UWORD		weight:5, processor: 27;
	
} DECREF_PACKET;

typedef struct GETID_PACKET
{
	WORD		number;						/* return the requested no of id's	*/
	
} GETID_PACKET;

typedef struct PUTID_PACKET
{
	/*	take care this packet does not take more than the maximum system
		packet size */
	
	WORD		number;						/* the number of id's				*/
	WORD		i[VariableArraySize];		/* the id's							*/
	
} PUTID_PACKET;

typedef struct ALLOCATION_PACKET
{
	GRAPH_ID	id;
	WORD		size;
	BOOL		reduce;						/* true if reducer must be allocated */
	
} ALLOCATION_PACKET;

typedef struct IN_PACKET
{
	GRAPH_ID	id;
	BOOL		reserved;					/* indicates success					*/
	BOOL		reduce;						/* true if reducer must be allocated	*/
	
} IN_PACKET;

typedef struct VOUT_BLOCK_PACKET
{   
	GRAPH_ID	id;
	WORD		size;						/* total size of the block */
	WORD		position;
	BYTE		data[VariableArraySize];	/* the contents of the block */

} VOUT_BLOCK_PACKET;

typedef struct VOUT_GRAPH_PACKET
{
	GRAPH_ID	id;
	WORD		size;						/* total size of the graph			*/
	ADDRESS		reduce;						/* startup code (if any)			*/
	WORD		position;
	BYTE		graph[VariableArraySize];

} VOUT_GRAPH_PACKET;

typedef struct OUT_BASIC_PACKET
{
	GRAPH_ID	id;
	BYTE		basic[BasicNodeSize];

}OUT_BASIC_PACKET;

/************************************************************************************
 *																					*
 *		heap packets																*
 *																					*
 ************************************************************************************/

typedef struct VBYTE_HPACKET
{
	PACKET_HEADER		header;
	WORD				source;
	BYTE				data[VariableArraySize];	/* the contents of the block */
	
} VBYTE_HPACKET;

typedef struct VBYTE_STRING_HPACKET
{
	PACKET_HEADER		header;
	WORD				source;
	VOUT_BLOCK_PACKET	packet;

} VSTRING_HPACKET;

typedef struct VBYTE_GRAPH_HPACKET
{
	PACKET_HEADER		header;
	WORD				source;
	VOUT_GRAPH_PACKET	packet;

} VGRAPH_HPACKET;

/********************************************************************************
 *																				*
 *		exported routines														*
 *																				*
 ********************************************************************************/

extern BOOL initialise_graph_manager(VOID);
extern VOID request_process(VOID);

#endif
