#ifndef _HARDWARE_
#define _HARDWARE_

#ifdef __INCLUDE_STANDARD__

#include <limits.h>
#include <par.h>
#include <asm.h>
#include <i64.h>

#endif

#define HighPrio			0
#define LowPrio				1

#define NLinks				4
#define VLinks				(NLinks+1)		/* virtual links: n links + 1 */
#define WordLength			(sizeof(int))

#define ByteSelectMask		3
#define ByteSelectLength	2
#define BPW					32			/* bits per word */

/********************************************************************************
 *																				*
 *		function prototypes														*
 *																				*
 ********************************************************************************/

extern void	*mint					(void);

/********************************************************************************/

extern void	*hard_stop_wsp			(void);	
extern void	|*hard_signal_address	(void);
extern void	*hard_root				(void);
extern void	*hard_descriptors		(void);

/********************************************************************************/

extern void	high					(void);
extern void	low						(void);

extern void	sthf					(const void *wsp);
extern void	sthb					(const void *wsp);
extern void	stlf					(const void *wsp);
extern void	stlb					(const void *wsp);
extern void	saveh					(const void *p);
extern void	savel					(const void *p);

extern void	prime_scheduler			(void);

extern void	stopp					(const void *wsp);
extern void	runp					(const void *wsp, const int pri);
extern void	move					(const void *to,const void *from,const int len);

/********************************************************************************/

extern void	|*link_to_chanout		(const int n);
extern void	|*link_to_chanin		(const int n);
extern void	|*event_channel			(void);

extern int	chanout_to_link			(const void |*ch);
extern int	chanin_to_link			(const void |*ch);

extern void	sendlink				(const int, const void *, const size_t);
extern void	recvlink				(const int, const void *, const size_t);

/*********************************************************************************/

extern void	wait_channel			(const void |*ch);
extern void signal_channel			(const void |*ch);

extern void	wait					(const int ticks);

/************************************************************************************
 *																					*
 *		arithmetic prototypes														*
 *																					*
 ************************************************************************************/

extern void	i64tor64				(const int *i,const double *r);

extern void	fpuabs					(const double *r);

extern void	fpint_nearest			(const double *r);
extern void	fpint_zero				(const double *r);
extern void fpint_plus_inf			(const double *r);
extern void fpint_min_inf			(const double *r);

extern int	minus					(const int a,const int b);
extern int	plus					(const int a,const int b);
extern int	after					(const int a,const int b);


/********************************************************************************
 *																				*
 *		definition of macro's													*
 *																				*
 ********************************************************************************/

#define	mint()					(((void * (*) (const char *)) __asm) ("mint"))

#define MostNeg					((int)mint())
#define NotProcessP				(mint())
#define UniqBase				((int)mint())

#define UsrMemStart				((void *)((int)mint()+0x70))
#define ExtMemStart				((void *)((int)mint()+0x1000))

/* addresses of special fields */

#define hard_WdescIntSaveLoc	(((void ** (*) (const char *)) __asm) ("mint,ldnlp 11"))
#define hard_IptrIntSaveLoc		(((void ** (*) (const char *)) __asm) ("mint,ldnlp 12"))
#define hard_AregIntSaveLoc		(((int * (*) (const char *)) __asm) ("mint,ldnlp 13"))
#define hard_BregIntSaveLoc		(((int * (*) (const char *)) __asm) ("mint,ldnlp 14"))
#define hard_CregIntSaveLoc		(((int * (*) (const char *)) __asm) ("mint,ldnlp 15"))

#define hard_IntWdesc			(((void * (*) (const char *)) __asm) ("mint,ldnl 11"))
#define hard_IntIptr			(((void * (*) (const char *)) __asm) ("mint,ldnl 12"))
#define hard_IntAreg			(((int (*) (const char *)) __asm) ("mint,ldnl 13"))
#define hard_IntBreg			(((int (*) (const char *)) __asm) ("mint,ldnl 14"))
#define hard_IntCreg			(((int (*) (const char *)) __asm) ("mint,ldnl 15"))

#define hard_IntWsp				((void *)((int)hard_IntWdesc&~ByteSelectMask))

#define	hard_stop_wsp()			(((void * (*) (const char *)) __asm) ("mint,ldnlp 28"))
#define	hard_signal_address()	(((void |* (*) (const char *)) __asm) ("mint,ldnlp 29"))

/* contents of special fields */

#define	hard_root()				(((void * (*) (const char *)) __asm) ("mint,ldnl 30"))
#define	hard_descriptors()		(((void * (*) (const char *)) __asm) ("mint,ldnl 31"))

/**********************************************************************************

	increase prio starts up a new high priority process just after setting its
	wsp to a special word in low memory. This works for both priorities. when
	a low prio process executes this sequence will be interrupted and the stopp
	will be executed after interruption. In case a high prio process executes this
	sequence the stopp will be executed directly. Note this routine depends on the
	size of instructions after ldpi (6*2)

	decrease prio just puts low prio process at the end of scheduling list and stops
	current process. Usable in any priority
	
************************************************************************************/

#define	high()					(__asm ("ldc 12,ldpi,stl -1,mint,ldnlp 28,gajw,runp,stopp"))
#define	low()					(__asm ("ldlp 0, adc 1,runp,stopp"))

#define	sthf(n)					(((void (*) (const char *, const void *)) __asm) ("sthf", (n)))
#define	sthb(n)					(((void (*) (const char *, const void *)) __asm) ("sthb", (n)))
#define	stlf(n)					(((void (*) (const char *, const void *)) __asm) ("stlf", (n)))
#define	stlb(n)					(((void (*) (const char *, const void *)) __asm) ("stlb", (n)))
#define	saveh(p)				(((void (*) (const char *, const void *)) __asm) ("saveh", (p)))
#define	savel(p)				(((void (*) (const char *, const void *)) __asm) ("savel", (p)))

#define prime_scheduler()		(__asm ("ldtimer,ldc 8,sttimer,ldc 0,sttimer,ldc 8,sttimer,ldc 0,sttimer,sttimer"))

#define stopp(wsp)				(((void (*) (const char *, const void *)) __asm) ("ldlp 0,rev,stnl 0,stopp", (wsp)))
#define runp(wsp,pri)			(((void (*) (const char *, const void *)) __asm) ("runp", (void *) ((int)(wsp)|(pri))))
#define move(to,from,len) 		(((void (*) (const char *, const int, const void *, const void *)) __asm) ("move", (len), (to), (from)))

/********************************************************************************/

#define	link_to_chanout(link)	((void |*)(WordLength*(int)(link)+(int)mint()))
#define	link_to_chanin(link)	((void |*)(WordLength*((int)(link)+NLinks)+(int)mint()))
#define	event_channel()			((void |*)(2*NLinks*WordLength+(int)mint()))

#define	chanout_to_link(ch)		(((int)(ch)-(int)mint())/WordLength)
#define	chanin_to_link(ch)		(((int)(ch)-(int)mint())/WordLength-NLinks)

#define	sendlink(link,buf,len)	chwrite(link_to_chanout(link),(buf),(len))
#define	recvlink(link,buf,len)	chread(link_to_chanin(link),(buf),(len))

/***********************************************************************************/

#define	wait_channel(ch)		|*(char |*)(ch)
#define	signal_channel(ch)		{if (*(int*)(ch)!=(int)mint()) |*(char |*)(ch) = '\0';}

#define	wait(ticks)				wttimer(ticks+ldtimer())

/************************************************************************************
 *																					*
 *		some special arithmetic	(low level)											*		
 *																					*
 ************************************************************************************/

/*
	stacktop:		used to pop values, use B=(A=function(),stacktop())
	
	i64inci32:		increase 64 bit integer with 32 bit integer
	i64tor64:		convert 64 bit integer to 64 bit floating point number
	
	fpuabs:			absolute value of 64 bit float
	fpint_...		rounded 64 bit float
	
	minus,plus,after: manipulation of cyclic quantities, such as time
*/

#define	stacktop()			(((int (*) (const char *)) __asm) (""))

#define i64incu32(n,i)		(n.hi=ladd(n.hi,0,(n.lo=lsum(n.lo,i,0).lo,stacktop())))
#define i64tor64(n,r)		(((void (*) (const char *,const int *,const double *)) __asm) ("dup,ldnlp 1,fpi32tor64,fpuexpinc32,fpb32tor64,fpadd,fpstnldb", ((int *)(n)),(r)))

#define fpuabs(r)			(((void (*) (const char *,const double *)) __asm) ("dup,fpldnldb,fpuabs,fpstnldb",(r)))

#define fpint_nearest(r)	(((void (*) (const char *,const double *)) __asm) ("dup,fpldnldb,fpurn,fpint,fpstnldb",(r)))
#define fpint_zero(r)		(((void (*) (const char *,const double *)) __asm) ("dup,fpldnldb,fpurz,fpint,fpstnldb",(r)))
#define fpint_plus_inf(r)	(((void (*) (const char *,const double *)) __asm) ("dup,fpldnldb,fpurp,fpint,fpstnldb",(r)))
#define fpint_min_inf(r)	(((void (*) (const char *,const double *)) __asm) ("dup,fpldnldb,fpurm,fpint,fpstnldb",(r)))

#define	minus(a,b)			(((int (*) (const char *,const int,const int)) __asm) ("diff", (b),(a)))
#define plus(a,b)			(((int (*) (const char *,const int,const int)) __asm) ("sum", (b),(a)))
#define after(a,b)			(minus(a,b)>0)

/************************************************************************************
 *																					*
 *		undefine some constants to avoid problems in stdlib.h						*
 *																					*
 ************************************************************************************/

#undef ldiv

#endif