#include "heapcopy.h"

#include "kernel.h"
#include "macros.h"
#include "depend.h"
#include "heapcollect.h"
#include "hardware.h"
#include "graphs.h"
#include "messageno.h"

#ifdef __INCLUDE_STANDARD__

#endif

extern PROCESSOR_INFO	processor_info;
extern NODE_INFO		node_info;
extern SYS				sys;

/********************************************************************************
 *																				*
 *		administration															*
 *																				*
 ********************************************************************************/

VOID add_copy_time (UWORD start_time)
{
	UWORD stop_time;
	
	stop_time=(ldtimer()-start_time);
	
	i64incu32(processor_info.ticks_copying,stop_time);
}

/********************************************************************************
 *																				*
 *		copying of graphs (at high prio)										*
 *		the heap is supposed to be in valid state and the system stack			*
 *		is supposed to be empty													*
 *																				*
 ********************************************************************************/

static NODE *try_copy_graph(NODE *graph)
{
	/* make a global copy of the argument graph. return the
	   resulting root node. To speed up all this, the copier can be placed in
	   on-chip memory temporarily. Another reference to the graph must exist
	   somewhere so that the garbage collector will not throw it away. after
	   a garbage collection has been performed this routine returns NULL, to
	   indicate it should be executed again: this is necessary as this routine
	   needs the new address of the graph after garbage collection. One could
	   get it from the forwarding address of the old node, but this depends
	   on the kind of collector used... */

	NODE *copy;
	WORD size;
	UWORD start_time;

#if __CACHE_ROUTINES__

	WORD copy_routine_size, restore_routine_size, max_routine_size;
	LABEL *routine;
	
	copy_routine_size		=(ADDRESS)copy__routine__end-(ADDRESS)CopyGraph;
	restore_routine_size	=(ADDRESS)restore__routine__end-(ADDRESS)RestoreGlobalGraph;

	max_routine_size=max(copy_routine_size,restore_routine_size);
	
#endif
		
	start_time=ldtimer();

#if __CACHE_ROUTINES__
	save_registers(max_routine_size);
	routine=routine_on_chip((LABEL *)CopyGraph,copy_routine_size);
	copy=(NODE *)(*routine)(graph);
#else
	copy=CopyGraph(graph);
#endif
	
	if((WORD)copy>=0)
	{
		/*	amount of heap needed is returned on failure, forget copy,
			restore original graph, call garbage collector and return NULL */

#if __CACHE_ROUTINES__
		restore_registers(max_routine_size);
#endif
		RestoreLocalGraph();
		add_copy_time(start_time);
			
		collect((WORD)copy);
		return NULL;
	}

	/* copy succeeded, restore graph */
	
#if __CACHE_ROUTINES__
	routine=routine_on_chip((LABEL *)RestoreGlobalGraph,restore_routine_size);
	size=(WORD)(*routine)();
	add_copy_time(start_time);
	restore_registers(max_routine_size);
#else
	size=RestoreGlobalGraph();
	add_copy_time(start_time);
#endif

	if(size>0)
	{
		/* restoring failed, amount of heap needed is returned */
		
		collect((copy->string_node.string->length>>2)+size+3);
		return NULL;
	}
	return copy;
}

VOID copy_saved_graph (WORD id)
{
	NODE *copy;
	
	copy=NULL;
	while(copy==NULL)
	{
		procresch();
		copy=try_copy_graph(table_graph(id));
	}
	
	table_graph(id)=copy;
}

VOID graph_message(WORD id, WORD destproc, WORD destid, ADDRESS code)
{
	NODE				*message;
	VOUT_GRAPH_PACKET	*out;
	
	message=alloc_hpacket2(MOutGraph,destproc,vsizeof(VOUT_GRAPH_PACKET)-sizeof(WORD),id,&out);
	
	out->id.dest		=destid;
	out->id.source		=id;
	out->size			=table_graph(id)->string_node.string->length;
	out->reduce			=code;
	
	table_graph(id)=message;
}


VOID block_message(WORD id, WORD destproc, WORD destid)
{
	NODE				*message;
	VOUT_BLOCK_PACKET	*out;
	
	message=alloc_hpacket2(MOutBlock,destproc,vsizeof(VOUT_BLOCK_PACKET)-sizeof(WORD),id,&out);
	
	out->id.dest		=destid;
	out->id.source		=id;
	out->size			=table_graph(id)->string_node.string->length;
	
	table_graph(id)=message;
}

VOID allocate_and_transmit(WORD id,WORD kind,WORD size,WORD reduce,WORD destproc, WORD destid)
{	
	ALLOCATION_PACKET *out;
	NODE *message;

	message=alloc_hpacket(kind,destproc,sizeof(ALLOCATION_PACKET),&out);
	
	out->id.dest			=destid;
	out->id.source			=id;
	out->size				=size;
	out->reduce				=reduce;
	
	send_heap_message(message);
}