/********************************************************************
 *																	*
 *		file server													*
 *																	*
 ********************************************************************/

#include "hostserver.h"

#ifndef _TYPES_
#include "types.h"
#endif

#include "hardware.h"
#include "messageno.h"
#include "hostio.h"
#include "remoteio.h"
#include "packets.h"
#include "router.h"
#include "loader.h"
#include "memory.h"
#include "commands.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

#endif

extern NODE_INFO		node_info;
extern NET_INFO			net_info;
extern PROCESSOR_INFO	processor_info;
extern OPTIONS			options;

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

static WORD			started;
static FLOAT		total_time;
static WORD			n_exited;

CHAR host_buffer[StringLength+1];

/********************************************************************************
 *																				*
 *		global debug variables													*
 *																				*
 ********************************************************************************/

/* the server filters some debug commands... */

typedef enum DEBUG_PRIVILEGE
{
	D_ENABLED,	D_DISABLED
	
} DEBUG_PRIVILEGE;

static VCOMMAND_TABLE *io_command_table;
static WORD *debug_privilege;

/********************************************************************************
 *																				*
 *		table of actions to be taken on reception of a packet					*
 *																				*
 ********************************************************************************/

static RECEIVE_ACTION *receive_table;

/********************************************************************************
 *																				*
 *		router interface														*
 *																				*
 ********************************************************************************/

static BOOL packet_available(VOID)
{
	return False;
}

static VOID output_packet(VPACKET_BUFFER *b)
{
	(*receive_table[b->header.kind])(b->header.kind,b->body.source,b->header.length-sizeof(WORD),(VPACKET *)&b->body.bytes);
}

static VOID input_packet(VPACKET_BUFFER *b)
{
	fprintf(stderr,"READ FROM HOST?\n");
}

/********************************************************************************
 *																				*
 *		exported routines														*
 *																				*
 ********************************************************************************/

VOID exit_ioserver(VOID)
{
	host_close_files();
	
	/* if there is a remote server running, close link to it */
	
	remote_disconnect();
	kill_router();
	
	exit(0);
}

/********************************************************************************
 *																				*
 *		packet handling routines												*
 *																				*
 ********************************************************************************/

/********************************************************************************
 *																				*
 *		miscelaneous routines													*
 *																				*
 ********************************************************************************/
 
static VOID handle_exit(WORD kind,WORD source,WORD size,VPACKET *packet)
{
	exit_ioserver();
}

static VOID reset_server_clock(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	started = ldtimer();
}

static VOID read_command(WORD kind,WORD source,WORD size,VCOMMAND_PACKET *packet)
{
	WORD	length;
	
	if (source < 0)
		source-=UniqBase;
	
	/* first check privilege of this transputer, if it is not allowed to do
	   anything, keep transputer pending */
	
	if (debug_privilege[source] == D_DISABLED)		
		return;

	/* transputer may issue command, so read it and execute */
	
	for (;;)
	{	
		fflush(stderr);

		/* write prompt */
		
		fprintf(stdout,"[%d]",source);
		host_write_string (stdout,packet->command.characters,packet->command.length);
		fflush(stdout);
	
		length=host_read_string(stdin,host_buffer,StringLength);
	
		if (!exec_command(io_command_table,host_buffer))
		{
			/* the command could not be filtered, so pass it to transputer and return */
			
			VCOMMAND_PACKET *return_packet;
	
			return_packet=alloc_packet(False);
			
			return_packet->command_dest			=packet->command_source;
			return_packet->command.length		=length;
			
			move(return_packet->command.characters,host_buffer,length);
		
			send_packet_from(MOutCommand,source,ToHost,vsizeof(VCOMMAND_PACKET)+length,return_packet,False);
			
			return;
		}
	}
}

static VOID host_command(WORD kind,WORD source,WORD size,VCOMMAND_PACKET *packet)
{
	VCOMMAND_PACKET *return_packet;
	WORD exit_status;
	
	if (source < 0)
		source-=UniqBase;
	
	/* the string already contains a termination NULL character */
		
	exit_status=system(packet->command.characters);

	return_packet=alloc_packet(False);
	
	return_packet->command_dest			=packet->command_source;
	return_packet->command.length		=exit_status;
	
	send_packet_from(MOutCommand,source,ToHost,vsizeof(VCOMMAND_PACKET),return_packet,False);
}

VOID program_ready(VOID)
{
	send_packet_from(MProgramReady,options.root_processor,ToHost,vsizeof(VPACKET),alloc_packet(False),False);		
}

VOID handle_exit2(WORD kind,WORD source,WORD size,PROCESSOR_INFO *info)
{
	FLOAT	garb_secs, active_secs, copy_secs;

	if (n_exited == 0)
	{
		/* this is the first exit message that has arrived, so fix total time */
	
		total_time=((FLOAT)minus(ldtimer(),started))/MHz;
		
		fflush(stdout);
		fflush(stderr);
	
		if (options.verbose)
			putc('\n',stdout);
	
		if (options.extra_verbose)	
			fprintf(stdout," proc:    copy: collect:  active:    idle:\n");
	}
	
	++n_exited;

	i64tor64(&info->ticks_collecting,	&garb_secs);
	garb_secs/=MHz;	
	i64tor64(&info->ticks_copying,		&copy_secs);
	copy_secs/=MHz;
	i64tor64(&info->ticks_running,		&active_secs);
	active_secs/=MHz;

	if (options.extra_verbose)
		fprintf(stdout,"%6d %8.3lf %8.3lf %8.3lf %8.3lf\n",info->id,copy_secs,garb_secs,active_secs,total_time-active_secs);
	
	
	if (n_exited >= net_info.size)
	{
		/* all processors have stopped, unlock the shell and release remote
		   server (if any) */
		
		if (options.verbose)
			fprintf(stdout,"total time: %0.3lf seconds.\n", total_time);
		
		/* flush all remote commands and leave connection open */
		
		remote_end();
		host_close_files();
		program_ready();
	}
}

static VOID handle_load (WORD kind,WORD source,WORD size,VCOMMAND_PACKET *packet)
{
	FILE				*file;
	
	VCODE_SEGMENT		*code_segment;
	WORD				buffer_size;
	UCHAR				*buffer;

	/* loading is done by means of synchronous messages. This means that
	   the output process of the server cannot handle any message
	   meanwhile... Be careful not to isue io requests in this time!.
	   On error as special position code is returned */

	/* initialise buffer space */
	
	code_segment=code_packet_address();		
	code_segment->position=0;

	buffer		=(UCHAR *)code_segment->code;
	buffer_size	=((UCHAR *)code_segment+options.packet_size)-buffer;

	/* try open program file */
	
	if (packet->command.length >= StringLength)
	{
		fprintf(stderr,"load: filename too long\n");
		
		code_segment->position=-2;
		send_packet_from(MProgramSegment,ToAll,ToHost,vsizeof(VCODE_SEGMENT),code_segment,True);
		return;
	}

	move(host_buffer,packet->command.characters,packet->command.length);
	host_buffer[packet->command.length]='\0';
	
	file = fopen (host_buffer,"rb");
	if (file == NULL)
	{
		fprintf(stderr,"file %s could not be opened\n",host_buffer);
		
		code_segment->position=-2;
		send_packet_from(MProgramSegment,ToAll,ToHost,vsizeof(VCODE_SEGMENT),code_segment,True);
		return;
	}

	/*	before loading:
		- clear remote screen
		- initialise file io
	*/

	remote_clear_screen();
	init_hostio();
	
	n_exited=0;
	
	for (;;)
	{
		/* fill the packet and send it to all transputers, including the
		   root, but not to the host */
		
		WORD n;
		
		n=fread(buffer,1,buffer_size,file);
		
		if (n < buffer_size)
		{
			/* this is the last packet */
			
			fclose(file);
			
			if (kind==MProgramLoad)
				code_segment->position=-1;
			else
				code_segment->position=-3;
				
			send_packet_from(MProgramSegment,ToAll,ToHost,vsizeof(VCODE_SEGMENT)+n,code_segment,True);
			
			remote_begin();
			
			return;
		}
		
		send_packet_from(MProgramSegment,ToAll,ToHost,options.packet_size,code_segment,True);
	}
}

/********************************************************************************
 *																				*
 *		server debug commands													*
 *																				*
 ********************************************************************************/

static VOID disable (BOOL available, WORD transputer)
{
	if (available)
		debug_privilege[transputer]=D_DISABLED;	
}

static VOID enable (BOOL available, WORD transputer)
{
	if (available)
		debug_privilege[transputer]=D_ENABLED;	
}

/********************************************************************************
 *																				*
 *		server connect commands													*
 *																				*
 ********************************************************************************/

static VOID connect (BOOL available, CHAR *name)
{
	if (available)
	{
		if (remote_connect(name))
		{
			remote_clear_screen();

			if (options.verbose)
				fprintf(stderr,"connection established\n");
		
		} else
		{
			fprintf(stderr,"connection failed\n");
			return;
		}	
	} else
		fprintf(stderr,"a server name was expected\n");
}

static VOID disconnect(VOID)
{
	if (remote_disconnect())
	{
		if (options.verbose)
			fprintf(stderr,"disconnected\n");
	
	} else
		fprintf(stderr,"could not disconnect\n");
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

typedef struct SERVER_ACTION
{
	WORD			number;
	RECEIVE_ACTION	action;
	
} SERVER_ACTION;

static SERVER_ACTION action_table[] =
{
	{MExit,				handle_exit},
	{MResetClock,		reset_server_clock},
	{MInCommand,		read_command},
	{MHostCommand,		host_command},
	{MRemoteServer,		handle_remote_server},

	{MProgramHalted,	handle_exit2},
	{MProgramHalt,		ignore_reception},
	{MProgramLoad,		handle_load},
	{MProgramDebugLoad,	handle_load},
	
	{MOutString,		host_output_string},
	{MOutCString,		host_output_c_string},
	{MInString,			host_input_string},
	{MInCString,		host_input_c_string},
	{MOutCharacter,		host_output_character},
	{MOutCCharacter,	host_output_c_character},
	{MInCharacter,		host_input_character},
	{MOutCInteger,		host_output_c_integer},
	{MOutCReal,			host_output_c_real},
	{MOutCBoolean,		host_output_c_boolean},
	
	{MOpenFile,			host_open_file},
	{MEndFile,			host_end_file},
	
	{MFatalError,		host_fatal_error},
	{MRuntimeError,		host_runtime_error},
	{MRuntimeWarning,	host_runtime_warning},
	
	{MReserveGraph,		host_handle_reserve},
	{MReserveBlock,		host_handle_reserve},
	
	{MErrno,			host_errno}
};

static COMMAND host_commands []=
{
	{"disable",		parse_n,		disable},
	{"enable",		parse_n,		enable},
	{"connect",		parse_word,		connect},
	{"disconnect",	parse_e,		disconnect}
};

BOOL init_ioserver(WORD n_message_kinds)
{
	/* to be called at the root transputer */
	
	WORD i;
		
	if(!(receive_table=push_mem(n_message_kinds*sizeof(RECEIVE_ACTION))))
		return False;
	if (!(io_command_table=get_command_table(10)))
		return False;
	if (!(debug_privilege=push_mem((net_info.size+1)*sizeof(WORD))))
		return;
	
	for(i=0;i<n_message_kinds;++i)
		receive_table[i]=host_default_reception;
	for(i=0;i<=net_info.size; ++i)
		debug_privilege[i]=D_ENABLED;

	/* set actions to be taken when packet arrives */
	
	for (i=0; i<sizeof(action_table)/sizeof(SERVER_ACTION); ++i)
		receive_table[action_table[i].number]=action_table[i].action;
	
	/* filtered debug commands */
	
	if (!define_commands(io_command_table,host_commands,sizeof(host_commands)/sizeof(COMMAND)))
		return False;
		
	if (!(start_virtual_link(chanout_to_link(node_info.to_host),output_packet,3*KiloByte,input_packet,KiloByte,packet_available)))
		return False;

	return initialise_remote_io();
}
