#ifndef _INSTRUCTIONS_
#define _INSTRUCTIONS_

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

/********************************************************************************
 *																				*
 *		instruction constants													*
 *																				*
 ********************************************************************************/

typedef enum OPERATOR
{
	Instr_unknown,
	
	Instr_pfix,		Instr_nfix,		Instr_opr,
	Instr_ldc,		Instr_ldl,		Instr_stl,		Instr_ldlp,
	Instr_adc,		Instr_eqc,		Instr_jump,		Instr_cjump,
	Instr_ldnl,		Instr_stnl,		Instr_ldnlp,
	Instr_call,		Instr_ajw,		Instr_lend,
	Instr_add,		Instr_sub,		Instr_mul,		Instr_div,		Instr_rem,
	Instr_sum,		Instr_diff,		Instr_prod,
	Instr_and,		Instr_or,		Instr_xor,		Instr_not,		Instr_shl,		Instr_shr,
	Instr_gt,
	Instr_bcnt,		Instr_wcnt,		Instr_ldpi,		Instr_mint,		Instr_bsub,		Instr_wsub,
	Instr_move,		Instr_in,		Instr_out,		Instr_lb,		Instr_sb,
	Instr_gcall,	Instr_gajw,		Instr_ret,		Instr_runp,		Instr_stopp,	Instr_ldtimer,
	Instr_dup,		Instr_rev,
	Instr_xdble,	Instr_xword,	Instr_ldiv,		Instr_lshr,		Instr_lshl,		Instr_ldpri,
	
	Instr_resetch,	Instr_sttimer,
	
	Instr_testerr,	Instr_fptesterr,
	
	Instr_sethalterr,
	
	Instr_outword,	Instr_outbyte,
	Instr_stlf,		Instr_sthf,		Instr_sthb,		Instr_stlb,

	Instr_fpldnldb,		Instr_fpstnldb,		Instr_fpi32tor64,
	Instr_fpadd,		Instr_fpsub,		Instr_fpmul,			Instr_fpdiv,
	Instr_fpdup,		Instr_fprev,		Instr_fpentry,
	Instr_fpgt,			Instr_fpeq,
	Instr_fpldzerodb,	Instr_fpint

} OPERATOR;

typedef enum ROUND_MODE
{
	ToNearest,
	ToZero,
	ToPlusInfinity,
	ToMinusInfinity
	
} ROUND_MODE;

/********************************************************************************
 *																				*
 *		instruction type														*
 *																				*
 ********************************************************************************/

typedef struct INSTRUCTION {
	
	BOOL		(*step_function) (PROCESSOR_STATE *processor_state, struct INSTRUCTION *instruction);
	
	OPERATOR	operator;
	WORD		arity;
	CHAR		*symbol;					/* shared, don't change */
	
	/* variables */
	
	WORD		operand;					/* value of Oreg */
	WORD		size;						/* size including prefices */
	ADDRESS		address;					/* address */

} INSTRUCTION;


/********************************************************************************
 *																				*
 *		exported routines														*
 *																				*
 ********************************************************************************/

extern BOOL initialise_instructions	(VOID);
extern VOID	set_round_mode			(ROUND_MODE mode);

/* low level address routines */

extern BOOL			test_word_address		(VOID *address);
extern BOOL			test_byte_address		(VOID *address);

extern ADDRESS 		get_instr_address		(ADDRESS address);
extern ADDRESS 		get_next_instr_address	(ADDRESS address);
extern ADDRESS 		get_prev_instr_address	(ADDRESS address);
extern ADDRESS 		skip_nops				(ADDRESS address);

/* low level instruction routines */

extern VOID			get_instr				(ADDRESS address, INSTRUCTION *instruction);
extern VOID			get_next_instr			(INSTRUCTION *instruction, INSTRUCTION *next_instruction);
extern CHAR			*get_instr_symbol		(INSTRUCTION *instruction);

extern BOOL			step					(INSTRUCTION *instruction);

/* low level mark routines */

extern ADDRESS 		next_mark				(ADDRESS address, WORD max);
extern ADDRESS		prev_mark				(ADDRESS address, WORD max);

/* data and symbols */

extern ADDRESS			get_data_area			(ADDRESS address);
extern ADDRESS			get_symbols				(ADDRESS data_area);
extern ADDRESS			get_symbols_and_data	(ADDRESS data_area, ADDRESS *symbols, ADDRESS *data, WORD *data_size);
extern ADDRESS			get_next_symbol			(ADDRESS symbol);
extern ADDRESS			get_symbol_definition	(ADDRESS data_area, ADDRESS symbol, CHAR **string);
extern ADDRESS			get_applied_symbol		(ADDRESS address, CHAR **string);
extern ADDRESS			get_data_area_end		(ADDRESS data_area);

/* high level address routines */

extern ADDRESS			get_offset_instr_address(ADDRESS address, WORD offset);

/* routines to show instructions */

extern ADDRESS		show_instruction		(WORD file_nr, ADDRESS iptr, ADDRESS address, ADDRESS data_area, ADDRESS *definition_symbol);
extern VOID			show_basic_instruction	(WORD file_nr, INSTRUCTION *instruction);
extern VOID			show_data				(WORD file_nr, ADDRESS data, WORD size);

extern VOID			show_code				(WORD file_nr, ADDRESS iptr, ADDRESS address, WORD count);

#endif
