#include "nodes.h"
#include "macros.h"

#ifndef _SYSTEM_
#include "system.h"
#endif

#include "io.h"

#ifdef __INCLUDE_STANDARD__
#include <string.h>
#endif

extern NODE_INFO	node_info;
extern SYS			sys;

NODE *skip_forwards(WORD file_nr, NODE *node)
{
	if (!test_word_address(node))
		return NULL;
	
	while (is_forward(node))
	{
		node=get_forward(node);
		
		if (!test_word_address(node))
			return NULL;
		
		file_print(file_nr,"->%08X: ",node);
	}
	return node;
}

WORD get_node_arity(NODE *node)
{
	ADDRESS code;
	
	code=get_code(node);
	
	if((WORD)code<0)
		return ((WORD)*(code+attr_arity_offset));
	
	return (desc_to_address(code)->arity);
}

static WORD get_node_type(NODE *node)
{
	ADDRESS code;
	
	code=get_code(node);
	
	if ((WORD)code<0)
		return ((WORD)*(code+attr_type_offset));
	return 0;
}

DESCRIPTOR *get_node_descriptor(NODE *node)
{
	ADDRESS code;
	
	code=get_code(node);
	
	if ((WORD)code<0)
	{
		WORD desc;
		
		desc=*((WORD *)code-2);
		
		if (desc & 2)
			return desc_to_address(desc-2);		/* strip mark... */
		else
		{
			runtime_warning(False,"closure without marked descriptor\n");
			return desc_to_address(desc);
		}
	}
		
	return desc_to_address(code);
}

WORD get_desc_arity(DESCRIPTOR *desc)
{
	WORD arity;
	
	arity=desc->arity;
	return (*((WORD *)desc-(arity<<1)-1));
}

VCLEAN_STRING *get_desc_symbol(DESCRIPTOR *desc)
{
	WORD desc_arity;
	WORD arity;
	
	arity=desc->arity;
	desc_arity=(*((WORD *)desc-(arity<<1)-1));
	
	return (VCLEAN_STRING *)((int *)desc-(arity<<1)+(desc_arity<<1)+2);
}

VOID show_symbol(WORD file_nr, VCLEAN_STRING *symbol)
{
	CHAR string[51];
	WORD length;
	
	length=symbol->length;
	
	if(length<0)
	{	file_print(file_nr,"symbol has negative length\n");
		return;
	}
	
	/* if an error occurred the symbol may be invalid: avoid printing 'infinite' strings */
	
	if(length>50)
		length=50;
	
	move(string,symbol->characters,length);
	string[length]='\0';
	
	file_print(file_nr,string);
}

static VOID show_remote_address (WORD file_nr, REMOTE_ADDRESS *address)
{
	file_print(file_nr,"[proc %d, index %d, weight %d]",address->processor,address->id,address->weight);
}

static VOID show_global_id(WORD file_nr, WORD id)
{
	if (id != 0)
		file_print(file_nr,"global node id: %d\n",id);

#if 0
	REQUEST_NODE *request_pointer;
	WORD length;
		
	if (wait_pointer<(WAIT_NODE *)0)
		file_print(file_nr,"local waitlist:\n");
	
	/*	only show limited amount of waiting local processes, so that wrong
		pointer don't result in a lot of garbage */
	
	for (length=0; length<50; length++)
	{
		if (wait_pointer >= (WAIT_NODE *)0)
			break;
	
		file_print(file_nr,"  pid: %d\n",wait_pointer->pid);		
		wait_pointer=wait_pointer->wait;
	}
	if (wait_pointer==NULL)
		return;
	
	request_pointer=(REQUEST_NODE *)wait_pointer;
	
	while (request_pointer > (REQUEST_NODE *)0)
	{
		/* an entry in global table found, search for requests */

		if ((WORD)request_pointer>=sys.comm_info.global_table_size)
		{	file_print(StdErr,"ERROR: node contains invalid table id: %d\n",request_pointer);
			return;
		}
		
		file_print(StdErr,"requests via id %d:\n",(WORD)request_pointer);

		request_pointer=sys.comm_info.global_table[(WORD)request_pointer].requests;
	
		while (request_pointer < (REQUEST_NODE *)0)
		{
			file_print(StdErr,"request from ");
			show_remote_address(file_nr,&request_pointer->address);
			file_print(StdErr,"\n");
			
			request_pointer=request_pointer->wait;
		}
		/* end of request list should point to another entry or nil */
	}
#endif
}

static VOID show_wait_list(WORD file_nr, NODE *node)
{
	for (;;)
	{
		WORD			arity;
		WORD			type;
		DESCRIPTOR		*desc;
		VCLEAN_STRING	*symbol;
		
		node=skip_forwards(file_nr,node);
		
		if (node == NULL)
			return;

		arity	=get_node_arity(node);
		type	=get_node_type(node);
		desc	=get_node_descriptor(node);
	
		if (arity<0)
		{	file_print(file_nr,"negative arity\n");
			return;
		}
		
		if (type & DeferedNode)
			file_print(StdErr,"D ");
		if (type & ReservedNode)
			file_print(StdErr,"R ");

		file_print(file_nr,"code:%08X ",get_code(node));

		symbol=get_desc_symbol(desc);
		show_symbol(file_nr,symbol);
		
		file_print(file_nr," pid: %d\n",node->release_node.pid);
		
		switch(arity)
		{
			case 0:
				if (node->release_node.wait != NULL)
					file_print(file_nr,"waitlist terminated by %08X",node->release_node.wait);
				return;
				
			case 1:
				node=(NODE *)node->release_node.wait;
				break;
			default:
				file_print(file_nr,"reserved node arity: %d\n",arity);
				show_node(file_nr,node);
				return;
		}
	}
}

static VOID show_hnf_node(WORD file_nr, NODE *node)
{
	DESCRIPTOR		*desc;
	WORD			desc_offset;
	WORD			arity;

	VCLEAN_STRING	*clean_string;
	
	arity=get_node_arity(node);
	desc=get_node_descriptor(node);
	desc_offset=(WORD)get_code(node);
	
	if(desc_offset==sys.descriptor_info.int_offset)
	{	file_print(StdErr,"INT %d\n", node->hnf0_node.basic);
		return;
	}
	
	if(desc_offset==sys.descriptor_info.bool_offset)
	{	if (node->hnf0_node.basic)
			file_print(file_nr,"BOOL TRUE\n");
		else
			file_print(file_nr,"BOOL FALSE\n");
		return;
	}
	
	if(desc_offset==sys.descriptor_info.char_offset)
	{	file_print(file_nr,"CHAR \'%c\'\n", node->hnf0_node.basic);
		return;
	}
	
	if(desc_offset==sys.descriptor_info.real_offset)
	{
		FLOAT			real;

		real = *(double *)&node->hnf0_node.basic;
		file_print(file_nr,"REAL %0.16e\n", real);
		return;
	}
	
	if(desc_offset==sys.descriptor_info.string2_offset)
	{
		CHAR			string[51];
		WORD			length;

		clean_string=(VCLEAN_STRING *)node->hnf0_node.basic;
		length=clean_string->length;
		
		if(length<0)
		{	file_print(file_nr,"string node with string of negative length %d\n",length);
			return;
		}
		
		if(length>50)
			length=50;
		
		move(string,clean_string->characters,length);
		string[length]='\0';
		
		if(length<clean_string->length)
			file_print(file_nr,"STRING \"%s...\" length: %d\n",string,clean_string->length);
		else
			file_print(file_nr,"STRING \"%s\" length: %d\n",string,clean_string->length);
		return;
	}
		
	if(desc_offset==sys.descriptor_info.file_offset)
	{	file_print(file_nr,"FILE (number %d, position %d)\n",node->file_node.number,node->file_node.position);
		return;
	}
	
	clean_string=get_desc_symbol(desc);
	show_symbol(file_nr,clean_string);
	
	switch (arity)
	{	case 0:
			file_print(file_nr,"\n");
			break;
		case 1:
			file_print(file_nr," %08X\n", node->hnf1_node.arg1);
			break;
		case 2:
			file_print(file_nr," %08X %08X\n", node->hnf2_node.arg1, node->hnf2_node.arg2);
			break;
		default:
		{
			NODE	**args;
			WORD	i;

			file_print(file_nr," %08X", node->hnf_node.arg1);
			args = node->hnf_node.arg_ptr;
			for (i=0; i<arity-1; i++)
				file_print(file_nr," %08X", args[i]);
			file_print(file_nr,"\n");
			break;
		}
	}
}

static VOID show_closure(WORD file_nr, NODE *node)
{
	WORD			arity;
	WORD			type;
	DESCRIPTOR		*desc;
	VCLEAN_STRING	*symbol;

	WORD			i;
	
	arity	=get_node_arity(node);
	type	=get_node_type(node);
	desc	=get_node_descriptor(node);

	if (arity<0)
	{	file_print(file_nr,"ERROR: closure has negative arity\n");
		return;
	}
	
	if (type & DeferedNode)
		file_print(StdErr,"D ");
	if (type & ReservedNode)
		file_print(StdErr,"R ");

	file_print(file_nr,"code:%08X ",get_code(node));
	
	symbol=get_desc_symbol(desc);
	show_symbol(file_nr,symbol);
	
	if (type & ChannelNode)
	{
		file_print(file_nr," ->");
		show_remote_address(file_nr,&node->channel_node.address);
	}

	if (type & ReservedNode)
	{
		file_print(file_nr,"\n");
		
		/*	see if this node has a waiting list, reserved nodes may
			only have arity 0 or 1 */
		
		switch(arity)
		{
			case 0:
				if (node->release_node.wait != NULL)
					file_print(file_nr,"waitlist terminated by %08X",node->release_node.wait);
				break;
			case 1:
				show_wait_list(file_nr,(NODE *)node->release_node.wait);
				break;
			default:
				file_print(file_nr,"reserved node arity: %d\n",arity);
		}
	} else
	{
		if(arity > 0)
		{	for (i=0; i<arity; i++)
				file_print(file_nr," %08X", node->closure.args[i]);
		}
	
		file_print(file_nr,"\n");
	}
}

VOID show_node(WORD file_nr, NODE *node)
{
	node=skip_forwards(file_nr,node);
	
	if (node == NULL)
		return;
		
	if(is_closure(node))		
		show_closure(file_nr,node);
	else		
		show_hnf_node(file_nr,node);
}
