#ifndef _NODES_
#define _NODES_

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

/********************************************************************************
 *																				*
 *		closure flags															*
 *																				*
 ********************************************************************************/

enum NODE_TYPES
{
	NormalNode		=0,
	DeferedNode		=1,			/* bit set if deferred entry			*/
	
	ReservedNode	=2,			/* bit set if entry for reserved node	*/
	
	Unused1			=4,			/* bit formerly used  for copy node		*/
	ChannelNode		=8			/* bit set if entry for channel node	*/
	
}NODE_TYPES;

enum ATTRIBUTE_OFFSETS
{
	attr_size_offset	=-4,
	attr_type_offset	=-3,
	attr_arity_offset	=-2
	
}ATTRIBUTE_OFFSETS;

/************************************************************************************
 *																					*
 *		rnf node types (descriptors are positive offsets from DATA descriptor)		*
 *																					*
 ************************************************************************************/

typedef struct DESCRIPTOR
{
	WORD	arity;
	WORD	lazy_offset;
	
} DESCRIPTOR;

typedef	struct STRING_NODE
{
	WORD			descriptor;
	VCLEAN_STRING	*string;
	
} STRING_NODE;

typedef struct REAL_NODE
{
	WORD		descriptor;
	FLOAT		real;
	
} REAL_NODE;

typedef struct FILE_NODE
{
	WORD		descriptor;
	WORD		number;
	WORD		position;
	
} FILE_NODE;

typedef struct BASIC_NODE
{
	WORD		descriptor;
	WORD		basic;
	
} BASIC_NODE;

typedef struct HNF0_NODE
{
		WORD		descriptor;
		WORD		basic;
		
} HNF0_NODE;
	
typedef struct HNF1_NODE
{
		WORD		descriptor;
		union NODE	*arg1;
		
} HNF1_NODE;
	
typedef struct HNF2_NODE
{
		WORD		descriptor;
		union NODE	*arg1;
		union NODE	*arg2;
		
} HNF2_NODE;
	
typedef struct HNF_NODE
{
		WORD		descriptor;
		union NODE	*arg1;
		union NODE	**arg_ptr;
		
} HNF_NODE;

/********************************************************************************
 *																				*
 *		closures																*
 *																				*
 ********************************************************************************/

typedef struct CLOSURE
{
		ADDRESS		code;
		union NODE	*args[1];
		
} CLOSURE;

typedef struct CHANNEL_NODE
{
	ADDRESS				code;
	struct CHANNEL_NODE	*next;			/* next active channel node */
	struct CHANNEL_NODE	*prev;			/* previous active channel node */
	REMOTE_ADDRESS		address;

} CHANNEL_NODE;

/********************************************************************************
 *																				*
 *		nodes with waiting lists (placeholders)									*
 *																				*
 ********************************************************************************/

typedef struct RELEASE_NODE
{
	ADDRESS					code;
	struct RELEASE_NODE		*wait;	
	WORD					pid;
	
} RELEASE_NODE;

typedef struct RESERVED_NODE
{
	ADDRESS					code;
	struct RELEASE_NODE		*wait;
	WORD					empty;
	
} RESERVED_NODE;

typedef struct REQUEST_NODE
{
	ADDRESS					code;
	struct REQUEST_NODE		*wait;
	REQUEST					request;
	
} REQUEST_NODE;

/********************************************************************************
 *																				*
 *		union of the above														*
 *																				*
 ********************************************************************************/

typedef union NODE
{
	STRING_NODE			string_node;
	REAL_NODE			real_node;
	FILE_NODE			file_node;
	BASIC_NODE			basic_node;
	CHANNEL_NODE		channel_node;
	
	HNF0_NODE			hnf0_node;
	HNF1_NODE			hnf1_node;
	HNF2_NODE			hnf2_node;
	HNF_NODE			hnf_node;
	
	CLOSURE				closure;
	
	RESERVED_NODE		reserved_node;
	REQUEST_NODE		request_node;	
	RELEASE_NODE		release_node;
	
} NODE;

#define BasicNodeSize	(sizeof(HNF2_NODE))

/********************************************************************************
 *																				*
 *		macros																	*
 *																				*
 ********************************************************************************/

#define get_code(node)			((ADDRESS)(node->closure.code))
#define get_forward(node)		((NODE *)((WORD)node->closure.code & (~3)))

#define is_forward(node)		(((WORD)get_code(node)&3)!=0)
#define is_closure(node)		((WORD)get_code(node)<0)
#define is_hnf(node)			((WORD)get_code(node)>=0)
#define is_nil(node)			((WORD)get_code(node)==0)

#define is_closure_type(node,type)	(*(((WORD *)get_code(node))-1)&(256*(type)))

#define desc_to_address(n)		((DESCRIPTOR *)(sys_descriptors+(WORD)(n)))
#define hnf_arity(node)			(desc_to_address(get_code(node))->arity)

#define fill_cons(n,a1,a2)		{n->hnf2_node.descriptor=sys.descriptor_info.cons_offset;n->hnf2_node.arg1=((NODE *)a1);n->hnf2_node.arg2=((NODE *)a2);}
#define fill_tuple2(n,a1,a2)	{n->hnf2_node.descriptor=sys.descriptor_info.tuple_offset+2*sizeof(DESCRIPTOR);n->hnf2_node.arg1=((NODE *)a1);n->hnf2_node.arg2=((NODE *)a2);}
#define fill_string(n,s)		{n->descriptor=sys.descriptor_info.string2_offset;n->string=((VCLEAN_STRING *)s);}

/********************************************************************************
 *																				*
 *		exported routines														*
 *																				*
 ********************************************************************************/

extern	WORD			get_node_arity			(NODE *node);
extern	DESCRIPTOR		*get_node_descriptor	(NODE *node);

extern	WORD			get_desc_arity			(DESCRIPTOR *desc);
extern	VCLEAN_STRING	*get_desc_symbol		(DESCRIPTOR *desc);

extern	VOID			show_symbol				(WORD file_nr, VCLEAN_STRING *symbol);
extern	VOID			show_node				(WORD file_nr, NODE *node);

#endif
