#ifndef _PACKETS_
#define _PACKETS_

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

/********************************************************************************
 *																				*
 *		macros																	*
 *																				*
 ********************************************************************************/

#define unique_rout(dest)	(UniqBase+(dest))

/********************************************************************************
 *																				*
 *		prototype of (system) packet types: packets may have arbitrary			*
 *		layout																	*
 *																				*
 ********************************************************************************/

typedef struct VPACKET
{
	BYTE	data[VariableArraySize];
	
} VPACKET;

typedef struct VCOMBINE_PACKET
{
	BYTE	packets[VariableArraySize];
	
} VCOMBINE_PACKET;

typedef struct VREDUCER_PACKET
{
	/* packet for some reducer */
	
	WORD	reducer_id;
	WORD	local_source;
	
	BYTE	contents[VariableArraySize];

} VREDUCER_PACKET;

typedef struct VCHANNEL_PACKET
{
	/* data for some process waiting on a channel */
	
	ADDRESS	destination;
	WORD	local_source;
	
	WORD	|*chan;
	BYTE	contents[VariableArraySize];

} VCHANNEL_PACKET;

/************************************************************************************
 *																					*
 *		local structure used by runtime system to link packets, normally this		*
 *		structure is hidden: the get_packet function can be used to get packets,	*
 *		which hides the list structure and the ready channel. However: the size		*
 *		of system packet is limited (to save memory) and thus they should only		*
 *		be used to send small messages. In some situations (e.g. program loading)	*
 *		one would like to have larger packets. This can be done by allocating a		*
 *		system packet of the right size explicitly. These packets should only be	*
 *		send in a synchronous way (which avoids overwriting of data that has not	*
 *		been sent yet). Furthermore the system_packet flag should be cleared to		*
 *		indicate this is not a standard system packet. The packet must be			*
 *		freed explicitly.															*
 *																					*
 ************************************************************************************/

typedef struct PACKET_HEADER
{
	WORD	size;
	WORD	kind;
	WORD	destination;
	
} PACKET_HEADER;

typedef struct QUEUE_ADMIN
{	
	struct QUEUE_PACKET		*next;
	BOOL					system_packet;
	CHANNEL					ready;
	
} QUEUE_ADMIN;

typedef struct VQUEUE_PACKET
{
	QUEUE_ADMIN				admin;
	PACKET_HEADER			header;
	WORD					source;
	BYTE					packet[VariableArraySize];
	
} VQUEUE_PACKET;

#define		queue_packet_offset				(sizeof(QUEUE_ADMIN)+sizeof(PACKET_HEADER)+sizeof(WORD))
#define		queue_address(packet_address)	((VQUEUE_PACKET *)((ADDRESS)(packet_address)-queue_packet_offset))
#define		packet_address(queue_address)	((VOID *)((ADDRESS)(queue_address)+queue_packet_offset))

/********************************************************************************
 *																				*
 *		routines to initialise packet communication								*
 *																				*
 ********************************************************************************/

extern BOOL initialise_packets			(WORD n_kinds);
extern BOOL start_packets				(VOID);

/********************************************************************************
 *																				*
 *		settings of packet router												*
 *																				*
 ********************************************************************************/

extern VOID block_heap_messages(VOID);
extern VOID unblock_heap_messages(VOID);

/********************************************************************************
 *																				*
 *		packet communication routines for system packets						*
 *																				*
 ********************************************************************************/

extern VOID				*alloc_packet		(BOOL sync);		/* out of small system pool				*/
extern VOID				*alloc_fatal_packet	(VOID);				/* prefab error packet					*/

extern VOID send_packet_from	(WORD kind,WORD dest,WORD from,WORD size,VOID *packet,BOOL sync);
extern VOID	send_fatal_packet	(WORD kind,WORD size,BOOL sync);

#define	send_packet(kind,dest,size,packet,sync)	send_packet_from((kind),(dest),(node_info.id),(size),(packet),(sync))

/********************************************************************************
 *																				*
 *		packet communication routines for heap packets							*
 *																				*
 ********************************************************************************/

extern NODE *alloc_hpacket_from		(WORD kind,WORD dest,WORD source,WORD size,VOID **packet_pointer);
extern NODE *alloc_hpacket2_from	(WORD kind,WORD dest,WORD source,WORD size,WORD body,VOID **header_pointer);

#define alloc_hpacket(kind,dest,size,packet)		alloc_hpacket_from(kind,dest,node_info.id,size,(VOID **)packet)
#define alloc_hpacket2(kind,dest,size,body,header)	alloc_hpacket2_from(kind,dest,node_info.id,size,body,(VOID **)header)

extern VOID	send_heap_message (NODE *message);

/********************************************************************************
 *																				*
 *		routines for packet handling											*
 *																				*
 ********************************************************************************/

#define ActivePacketSet		0
#define InactivePacketSet	1
#define	NPacketSets			2

typedef VOID (*RECEIVE_ACTION)(WORD kind,WORD source,WORD size,VOID *packet);
typedef RECEIVE_ACTION	RECEIVE_ACTIONS[NPacketSets];

typedef struct RECEIVE_ACTION_TABLE
{
	WORD				code;
	RECEIVE_ACTIONS		actions;
	
} RECEIVE_ACTION_TABLE;

/* the following routine should be used before 'start_packets' is issued, it sets
   the action to perform at reception of packet of certain kind */

extern VOID				set_packet_actions	(RECEIVE_ACTION_TABLE *actions,WORD amount);

extern RECEIVE_ACTION	get_packet_action	(WORD kind);
extern VOID				switch_action_set	(WORD set);

extern VOID				ignore_reception(WORD kind,WORD source,WORD size,VOID *packet);

#endif
