
#ifndef _PROCESSES_
#define _PROCESSES_

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

/********************************************************************************
 *																				*
 *		process status															*
 *																				*
 ********************************************************************************/

/*	it is not necessarily problematic if a suspended reducer has its status
	bit cleared, the scheduler will not swap it out in this case: it is not
	recognized as a suspended process, and it is also not in the transputer
	scheduling list. This property is used while initialising a process; we
	can assume that the locals of an on-chip reducer stay at a fixed place.
	
	it is NOT correct to set the waiting bit in a running process: it will
	incorrectly assume its id is positive, and the scheduler may crash, as
	it will encounter the same process multiple times.
*/

typedef enum PROCESS_STATUS
{
	PActive			=0,			/* process is running */
	PWaiting		=1			/* process is waiting */
	
} PROCESS_STATUS;

/************************************************************************************
 *																					*
 *		low level macro's (on workspaces)											*
 *																					*
 ************************************************************************************/

/*	the pointer to the next process is (wsp-2) is hardware-dependend,
	the prev offset is not. */

#define next_offset	(-2)
#define prev_offset	(-3)

#define set_next_process(this_wsp,next_wsp)		((VOID *)(*((WORD *)(this_wsp)+next_offset)=(WORD)(next_wsp)))
#define set_prev_process(this_wsp,prev_wsp)		((VOID *)(*((WORD *)(this_wsp)+prev_offset)=(WORD)(prev_wsp)))
#define set_iptr(wsp,code)						(*((WORD *)(wsp)-1)=(WORD)code)

#define get_next_process(wsp)					((VOID *)*((WORD *)(wsp)+next_offset))
#define get_prev_process(wsp)					((VOID *)*((WORD *)(wsp)+prev_offset))
#define get_iptr(wsp)							((ADDRESS)*((WORD *)(wsp)-1))

#define get_process_env(wsp)					((REDUCER_ENVIRONMENT *)((ADDRESS)(wsp)->heap_start-sizeof(struct REDUCER_ENVIRONMENT)))

#define move_idle_process(to,from)				move(&(to)->b_stack_ptr,&(from)->b_stack_ptr,4*sizeof(WORD))
#define move_process(to,from)					move((WORD *)(to)-1,(WORD *)(from)-1,sizeof(WORD)+sizeof(WSP))

/************************************************************************************
 *																					*
 *		low level macros (on environments)											*
 *																					*
 ************************************************************************************/

#define inactive_mask							((WORD)mint())

#define set_active(env)							((env)->pid&=~inactive_mask)
#define set_inactive(env)						((env)->pid|=inactive_mask)
#define is_active(env)							((env)->pid>=0)
#define is_inactive(env)						((env)->pid<0)
#define get_pid(env)							((env)->pid&~inactive_mask)

/************************************************************************************
 *																					*
 *		low level macros (on pids)													*
 *																					*
 ************************************************************************************/

#define pid_to_env(pid)							(sys.comm_info.id_table[pid].process.env)
#define pid_to_regs(pid)						((pid_to_env(pid))->registers)
#define	env_to_regs(env)						(env->registers)

/************************************************************************************
 *																					*
 *		low level routines															*
 *																					*
 ************************************************************************************/

extern VOID off_chip_suspend	(WSP *regs);
extern VOID off_chip_unsuspend	(WSP *regs);
extern VOID off_chip_activate	(WSP *regs);

/************************************************************************************
 *																					*
 *		reducer management I														*
 *																					*
 ************************************************************************************/

typedef VOID (*REDUCER_START_ROUTINE) (WORD pid);

extern WSP *init_reducer			(WSP *regs);

extern REDUCER_START_ROUTINE		reducer_start_routine;
#define set_restart_routine(r)		(reducer_start_routine=r)
#define get_restart_routine()		(reducer_start_routine)
#define run_reducer(pid)			((*reducer_start_routine)(pid))

extern VOID set_debug_run			(BOOL debug);

extern VOID simulate_run_reducer	(WORD pid);
extern VOID invalid_run_reducer		(WORD pid);

/************************************************************************************
 *																					*
 *		reducer management II														*
 *																					*
 ************************************************************************************/

/************************************************************************************
 *																					*
 *		scheduling process etc. (kernel stuff)										*
 *																					*
 ************************************************************************************/
 
extern VOID kill_reducers			(VOID);
extern VOID scheduler				(VOID);

#endif
