#include "scanner.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#endif

/********************************************************************************
 *																				*
 *		the scanner, basic scanning routines									*
 *																				*
 ********************************************************************************/

static BOOL alpha_num_table[257];			/* table with True for ident chars, False otherwise */

static VOID initialise_alpha_num_table(VOID)
{
	WORD i;
	
	UCHAR extra_alpha_nums[]="@#$%^&*-+/=<>\\_~.?\'\"";
	
	for (i=0; i<257; ++i)
		alpha_num_table[i]=False;	
	for (i='a'; i<='z'; ++i)
		alpha_num_table[i+1]=True;
	for (i='A'; i<='Z'; ++i)
		alpha_num_table[1+i]=True;
	for (i='0'; i<='9'; ++i)
		alpha_num_table[1+i]=True;
	for (i=0; i<sizeof (extra_alpha_nums); ++i)
		alpha_num_table[1+extra_alpha_nums[i]]=True;
}

#define is_ident_char(c) (alpha_num_table[(c)+1])

/* unlike some scanning algorithms for sequential machines these algorithms
   don't use global variables to keep track of characters and strings. This
   is important as context switches are possible at any moment */

#define next_char(string_handle)	(*string_handle)++
#define is_white(string_handle)		((**string_handle)==' '||(**string_handle)=='\t'||(**string_handle)=='\n')
#define skip_whites(string_handle)	{while(**string_handle==' '||**string_handle=='\t')next_char(string_handle);}

static VOID skip_to_char(CHAR **string_handle, CHAR c)
{
	for (;(c!=**string_handle)&&(**string_handle);next_char(string_handle));
}

#define skip_rest(string_handle) skip_to_char(string_handle,'\n')

static BOOL scan_hexa_digit(CHAR **string_handle, WORD *value)
{
	WORD digit;
	
	digit=(WORD)**string_handle;
	
	if (digit>='0' && digit<='9')
	{	*value=(digit-'0');
		next_char(string_handle);
		return True;
	}
	if (digit>='a' && digit<='f')
	{	*value=(digit-'a'+10);
		next_char(string_handle);
		return True;
	}
	if (digit>='A' && digit<='F')
	{	*value=(digit-'A'+10);
		next_char(string_handle);
		return True;
	}
	return False;
}

static BOOL scan_sign(CHAR **string_handle, WORD *sign)
{
	skip_whites(string_handle);
	if (**string_handle=='-')
	{	next_char(string_handle);
		*sign=-1;
		return True;
	}
	if (**string_handle=='+')
	{	next_char(string_handle);
		*sign=1;
		return True;
	}
	return False;
}

static BOOL scan_separator(CHAR **string_handle)
{
	skip_whites(string_handle);
	if (**string_handle==',')
	{	next_char(string_handle);
		return True;
	}
	return False;
}

static BOOL scan_BS_char(CHAR **string_handle, WORD *c)
{
	WORD n;
	WORD character;
	
	n=0;
	next_char(string_handle);					/* skip backslash */
	
	character=**string_handle;
	switch (character)
	{	case 'n':	*c='\n'; next_char(string_handle); return True;
		case 't':	*c='\t'; next_char(string_handle); return True;
		case 'r':	*c='\r'; next_char(string_handle); return True;
		case 'f':	*c='\f'; next_char(string_handle); return True;
		default:
			if (isdigit (character))
			{	n = 100 * (character - '0');
				next_char(string_handle);
				character=**string_handle;
				if (isdigit (character))
				{	n += 10 * (character - '0');
					next_char(string_handle);
					character=**string_handle;
					if (isdigit (character))
					{	*c = n + character - '0';
						next_char(string_handle);
						return True;
					}
					else
						return False;
				}
				else
					return False;
			}
			else
			{	*c=character;
				next_char(string_handle);
				return True;
			}
	}
}

/********************************************************************************
 *																				*
 *		exported scanning routines												*
 *																				*
 ********************************************************************************/

BOOL scan_character(CHAR **string_handle, WORD *character)
{
	skip_whites(string_handle);
	if(**string_handle!='\'')
	{	*character='\0';
		return False;
	}
	
	next_char(string_handle);
	if (**string_handle=='\\')
	{	if(!scan_BS_char(string_handle,character))
			return False;
	}else
	{	*character=**string_handle;
		next_char(string_handle);
	}
	
	if(**string_handle!='\'')
		return False;
	
	next_char(string_handle);
	return True;
}

BOOL scan_string(CHAR **string_handle, CHAR *string, WORD max)
{
	WORD curpos;
	WORD character;
	
	curpos=0;
	skip_whites(string_handle);
		
	if(**string_handle!='\"')
	{	string[0]='\0';
		return False;
	}
	
	next_char(string_handle);
	character=**string_handle;
	while(character!='\"')	
	{	if (character == '\\')
		{	if(scan_BS_char(string_handle,&character))
				string[curpos]=character;
			else
				return False;
		} else			
		{	string[curpos]=character;
			next_char(string_handle);
		}
		character=**string_handle;
		if((++curpos) >max)
		{	string[--curpos]='\0';
			return False;
		}
	}
	next_char(string_handle);
	string[curpos]='\0';
	return True;
}

BOOL scan_word(CHAR **string_handle, CHAR *word, WORD max)
{
	WORD curpos;
	
	curpos=0;
	skip_whites(string_handle);
	
	while(!is_white(string_handle))	
	{	word[curpos]=**string_handle;
		next_char(string_handle);
		if((++curpos) >max)
		{	word[--curpos]='\0';
			return False;
		}
	}
	word[curpos]='\0';
	return (curpos!=0);
}

BOOL scan_ident(CHAR **string_handle, CHAR *ident, WORD max)
{
	WORD curpos;
	WORD character;
	
	curpos=0;
	skip_whites(string_handle);
	
	if (isdigit(**string_handle))
	{	ident[0]='\0';
		return False;
	}
	
	character=**string_handle;
	while(is_ident_char(character))	
	{	ident[curpos]=character;
		next_char(string_handle);
		character=**string_handle;
		if((++curpos) >max)
		{	ident[--curpos]='\0';
			return False;
		}
	}
	ident[curpos]='\0';
	return (curpos!=0);
}

#define MAX_NUM_SIZE 10

BOOL scan_decimal(CHAR **string_handle, WORD *n)
{
	WORD digs;
	WORD character;
	
	skip_whites(string_handle);
	
	character=**string_handle;
	if (character<'0' || character>'9')
		return False;
	
	digs=1;
	*n=character-'0';
	
	next_char(string_handle);
	character=**string_handle;
	while (character>='0' && character<='9')
	{	if((digs++) > MAX_NUM_SIZE)
			return False;
		
		*n=*n * 10 + (character -'0');
		next_char(string_handle);
		character=**string_handle;
	}
	switch (character)
	{	case 'k':
		case 'K':
					*n = (*n)<<10;
					next_char(string_handle);
					break;
		case 'm':
		case 'M':
					*n = (*n)<<20;
					next_char(string_handle);
					break;
		default:
					break;
	}
	return True;
}

BOOL scan_hexa(CHAR **string_handle, WORD *n)
{
	WORD digit;
	
	skip_whites(string_handle);
	if (!scan_hexa_digit(string_handle,&digit))
		return False;
	
	*n = digit;
	while (scan_hexa_digit(string_handle,&digit))
		*n=(*n<<4) | digit;

	return True;
}

BOOL scan_positive_value(CHAR **string_handle, WORD *n)
{	
	skip_whites(string_handle);
	
	if(isdigit(**string_handle))
		return scan_decimal(string_handle,n);
	
	if(**string_handle=='$')
	{	next_char(string_handle);
		return scan_hexa(string_handle,n);
	}
	
	if (**string_handle=='\'')
		return scan_character(string_handle,n);
	
	return False;
}

BOOL scan_value(CHAR **string_handle, WORD *n)
{
	WORD sign;
	
	if(!scan_sign(string_handle,&sign))
		sign=1;
	
	if(!scan_positive_value(string_handle,n))
		return False;
	
	if(sign<0)
		*n = 0 - (*n);
		
	return True;
}

static WORD count_digits (WORD n)
{
	WORD count;
	
	for (count = 0; n > 0; count++)
		n = n / 10;
	return count;
}

static FLOAT exp10 (WORD n)
{
	FLOAT d;

	for (d = 1; n > 0; n--)
		d *= 10;
	return d;
}


BOOL scan_real (CHAR **string_handle, FLOAT *r)
{
	WORD sign;
	WORD signE;
	
	WORD u1,u2,u3,order;
	
	if(!scan_sign(string_handle,&sign))
		sign=1;
	
	/* a real is u1.u2Eu3 */
	
	if (!scan_decimal(string_handle,&u1))
		return False;
	
	if (**string_handle=='.')
		next_char(string_handle);
	else
		return False;

	for (order=0;**string_handle=='0'; order++)
		next_char(string_handle);
	
	if (!scan_decimal(string_handle,&u2))
	{	if(order > 0)
			u2 = 0;
		else
			return False;
	}
	order += count_digits(u2);
	
	if (**string_handle == 'E' || **string_handle == 'e')
	{	next_char(string_handle);
		if(!scan_sign(string_handle,&signE))
			signE=1;
		if (!scan_decimal(string_handle,&u3))
			return False;
	}
	else
		u3 = 0;

	if (signE>0)
		*r = ((FLOAT)u1 + (FLOAT)u2 / exp10(order)) * exp10(u3);
	else
		*r = ((FLOAT)u1 + (FLOAT)u2 / exp10(order)) / exp10(u3);
	
	if (sign<0)
		*r = 0 - (*r);
	
	return True;
}

BOOL initialise_scanner(VOID)
{
	initialise_alpha_num_table();
	return True;
}