#ifndef _SYSTEM_
#define _SYSTEM_

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _NODES_
#include "nodes.h"
#endif

#ifndef _HARDWARE_
#include "hardware.h"
#endif

/********************************************************************************
 *																				*
 *		signal layout															*
 *																				*
 ********************************************************************************/

typedef struct SIGNUM
{
	WORD		number;					/* signal number		*/
	struct WSP	*wsp;					/* wdesc: wsp + prio	*/
	
} SIGNUM;

/********************************************************************************
 *																				*
 *		processor state (used for debugging)									*
 *																				*
 ********************************************************************************/

typedef struct PROCESSOR_STATE
{
	/************************************************************************
	 *																		*
	 *		fields that should be initialised by caller						*
	 *		(more or less the local state of a process)						*
	 *																		*
	 ************************************************************************/

	WORD	Areg;
	WORD	Breg;
	WORD	Creg;
	
	FLOAT	fAreg;
	FLOAT	fBreg;
	FLOAT	fCreg;
	
	ADDRESS	Wptr;
	ADDRESS	Iptr;

	WORD	prio;
	
	BOOL	error_flag;
	BOOL	float_error_flag;
	
	/************************************************************************
	 *																		*
	 *		fields that are initialised by debugger							*
	 *		(more or less global state)										*
	 *																		*
	 ************************************************************************/
	
	WORD	time;					/* high prio time					*/
	WORD	round_mode;				/* set to ToNearest by default!!	*/
	
	ADDRESS	low_front;				/* only low prio processes are suspended on interrupt! */
	ADDRESS	low_back;

	/* start of special RAM locations */
	
	WORD	link0_out;
	WORD	link1_out;
	WORD	link2_out;
	WORD	link3_out;
	
	WORD	link0_in;
	WORD	link1_in;
	WORD	link2_in;
	WORD	link3_in;
	
	WORD	event_in;
	
	ADDRESS	high_timer_list;
	ADDRESS	low_timer_list;

	ADDRESS	WdescIntSaveLoc;
	ADDRESS	IptrIntSaveLoc;
	WORD	AregIntSaveLoc;
	WORD	BregIntSaveLoc;
	WORD	CregIntSaveLoc;
	
	WORD	STATUSIntSaveLoc;
	WORD	EregIntSaveLoc;
	
	/* end of special RAM locations */

} PROCESSOR_STATE;

typedef struct DEBUGGER_DATA
{		
	SIGNUM				|signal_channel;		/* channel to invoke debugger			*/
	PROCESSOR_STATE		*processor_state;		/* record with saved processor state	*/
	
} DEBUGGER_DATA;

/********************************************************************************
 *																				*
 *		info about other processors												*
 *																				*
 ********************************************************************************/

typedef struct DESTINATION_INFO
{
	WORD			proc_info;			/* index to information related to proc		*/
	WORD			distance_info;		/* index to information related to distance	*/
	
	WORD			proc;				/* record of proc						*/
	WORD			distance;			/* distance to proc						*/
	WORD			outlinks;			/* possible links out to proc			*/
	
	/* note the proc acts as source for the following field						*/
	
	WORD			outblinks;			/* links from proc to all others		*/

} DESTINATION_INFO;

/********************************************************************************
 *																				*
 *		process workspace layout												*
 *																				*
 ********************************************************************************/

typedef struct WSP {
	
	WORD				return_address;
	WORD				b1;
	WORD				b2;
	WORD				b3;
	WORD				b4;
	WORD				b5;
	WORD				b6;
	union NODE			*a1;
	union NODE			*a2;
	union NODE			*a3;
	union NODE			*a4;
	WORD				extra_local;
	
	WORD				*b_stack_ptr;
	union NODE			**a_stack_ptr;		/* points to virtual locals (on stack) as well */
	
	WORD				*heap_start;		/* points to end of environment as well (mint if empty) */
	WORD				*heap_end;
			
} WSP;

typedef struct HARD_WSP{
	
	WORD	wait_time;
	WORD	time_next_stat;
	
	WORD	comm_chan_ptr;
	WSP		*next_proc;
	ADDRESS	instr_ptr;
	
} HARD_WSP;

typedef	struct REDUCER_LOCALS
{
	HARD_WSP	special;
	WSP			wsp;
	
} REDUCER_LOCALS;

#define ReducerSize	(sizeof(REDUCER_LOCALS)/WordLength)		/* in words */

typedef struct REDUCER_ENVIRONMENT {

	WSP			*registers;				/* registers of reducer */
	WORD		pid;					/* highest bit indicates status (1=suspended) */

	union NODE	**a_stack_start;		/* bounds of stacks */
	WORD		*b_stack_start;
		
} REDUCER_ENVIRONMENT;

#define EnvironSize	(sizeof(REDUCER_ENVIRONMENT)/WordLength)			/* in words */

/********************************************************************************
 *																				*
 *		indirection table entry													*
 *																				*
 ********************************************************************************/

#define	RefFree			((UWORD)0)			/* free entry				*/
#define RefNull			((UWORD)-1)			/* null entry (index 0)		*/
#define RefProcess		((UWORD)-2)			/* process entry			*/
#define	RefAllocFail	((UWORD)-3)			/* special					*/
#define	RefIdBuff		((UWORD)-4)			/* id buffer				*/
#define RefMax			((UWORD)MostNeg)	/* 1..RefMax: graph entry	*/

typedef struct FREE_ENTRY
{
	UWORD				refnum;				/* RefFree			*/
	WORD				next;				/* next free entry	*/
	
} FREE_ENTRY;

typedef struct NULL_ENTRY
{
	UWORD				refnum;				/* RefNull	*/
	WORD				*p1;				/* NULL		*/
	WORD				*p2;				/* NULL		*/
	WORD				*p3;				/* NULL		*/
	WORD				i;					/* 0		*/

} NULL_ENTRY;

typedef struct PROCESS_ENTRY
{
	UWORD				refnum;				/* RefProcess				*/
	REDUCER_ENVIRONMENT	*env;				/* environment of reducer	*/
	WORD				*p1;				/* NULL						*/
	WORD				*p2;				/* NULL						*/
	WORD				i;					/* 0						*/

} PROCESS_ENTRY;

typedef struct GRAPH_ENTRY
{
	UWORD				refnum;				/* ref count between 1 and RefMax			*/
	union NODE			*graph;				/* graph that entry points to				*/
	union NODE			*reserved;			/* reserved area							*/
	struct REQUEST_NODE	*requests;			/* requests for graph						*/
	WORD				pid;				/* pid of reducer that reduces the graph	*/
	
} GRAPH_ENTRY;

typedef struct ALLOC_FAILED_ENTRY
{
	UWORD				refnum;				/* RefAllocFail				*/
	WORD				*p1;				/* NULL						*/
	union NODE			*reserved;			/* NULL	(mandatory)			*/
	WORD				*p3;				/* NULL						*/
	WORD				discarded;			/* no of discarded bytes	*/
	
} ALLOC_FAILED_ENTRY;

typedef struct IDBUFF_ENTRY
{
	UWORD				refnum;				/* RefIdBuff				*/
	union NODE			*buff1;				/* buffer					*/
	union NODE			*buff2;				/* buffer					*/
	struct RELEASE_NODE	*wait;				/* waiting processes		*/
	WORD				i;					/* 0						*/
	
} IDBUFF_ENTRY;

typedef union ENTRY
{
	FREE_ENTRY			free;
	NULL_ENTRY			null;
	PROCESS_ENTRY		process;
	GRAPH_ENTRY			graph;
	ALLOC_FAILED_ENTRY	alloc_fail;
	IDBUFF_ENTRY		idbuff;

} ENTRY;

/********************************************************************************
 *																				*
 *		sub-structures of the runtine system									*
 *																				*
 ********************************************************************************/

typedef struct COMM_INFO
{
	union NODE			*first_message;			/* first message in list */
	union NODE			*last_message;			/* last message in list */
	
	ENTRY				*id_table;			/* ptr to table with global entries */
	WORD				id_table_size;		/* nr of entries in table */
	WORD				id_table_free;		/* index of first free entry */
	
	WORD				transputer_id;			/* id of this transputer */
	WORD				n_transputers;			/* nr of transputers in network */
	
	/* three words below not yet initialised */
	
	WORD				*rout;					/* rout vectors for normal pnt-to-pnt routing */
	WORD				*urout;					/* rout vectors for unique path routing */
	WORD				*brout;					/* rout vectors for broadcast routing */
	
	DESTINATION_INFO	*dest_info;
	
	struct CHANNEL_NODE		*active_channels;			/* ptr to list of valid channel nodes */
	struct CHANNEL_NODE		*first_inactive_channel;	/* ptr to list of channels for which a decrement message should be sent */
	struct CHANNEL_NODE		*last_inactive_channel;		/* ptr to end of inactive channel list */

} COMM_INFO;

typedef struct REDUCER_INFO {

	WORD	init_stack_size;
	WORD	init_heap_size;
	WORD	ext_heap_size;

	WSP		*on_chip_free_list;			/* list of free on-chip regs */
	WSP		*on_chip_first_regs;		/* first on-chip reg (fixed) */
	WSP 	*on_chip_last_regs;			/* last on-chip reg (fixed)  */
	
	WSP		*off_chip_free_list;        /* list of free off-chip regs */
	WSP		*off_chip_suspended_regs;   /* list of suspended off-chip regs */
	WSP		*off_chip_first_regs;       /* first running off-chip reg */
	WSP		*off_chip_last_regs;		/* last running off-chip reg */
	
	WSP		*swap;						/* used for swapping regs */
		
} REDUCER_INFO;

typedef struct GENERAL_INFO {

	UWORD		runtime_flags;
	WORD		random_rand;
	WORD		random_m;
	WORD		random_a;
	WORD		random_b;
	
} GENERAL_INFO;

typedef struct COPY_INFO {

	WORD	return_workspace;
	WORD	words_needed;
	
	WORD	first_message;
	WORD	copy_first_env;
	WORD	copy_last_env;
	WORD	copy_original_nodes;
	WORD	copy_root_closure;
		
} COPY_INFO;

typedef struct DESC_INFO
{
	WORD	nil_offset;
	
	WORD	string1_offset;
	WORD	string2_offset;
	
	WORD	file_offset;
	WORD	real_offset;
	
	WORD	int_offset;
	WORD	bool_offset;
	WORD	char_offset;
	
	WORD	tuple_offset;
	WORD	cons_offset;
	WORD	if_offset;
	
} DESC_INFO;

typedef struct CODE_INFO
{
	ADDRESS		start_address;			/* runtime system will start first reducer here */

	/* reserved nodes */
	
	ADDRESS		reserve_code;			/* reserved nodes (without pid or gid) */
	ADDRESS		reserve1_code;
	
	/* request nodes */
	
	ADDRESS		request_code;			/* request list for certain graph */
	ADDRESS		request1_code;
	
	/* release nodes */
	
	ADDRESS		release_proc;			/* release nodes: contain pid or gid */
	ADDRESS		release_proc1;
	ADDRESS		release_request;
	ADDRESS		release_request1;
	
	ADDRESS		channel_code;			/* needed to construct channel */
	
	ADDRESS		ea_process_code;
	ADDRESS		ea_reduce_code;
		
} CODE_INFO;

/********************************************************************************
 *																				*
 *		main runtime system structure											*
 *																				*
 ********************************************************************************/

typedef struct ROOT
{
	ADDRESS				system_stack;
	ADDRESS				process_stack;

	ADDRESS				heap_free_ptr;
	ADDRESS				heap_end_ptr;
	ADDRESS				current_heap;
	ADDRESS				garbage_heap;
	WORD				half_heap_size;				/* in bytes */
	
	COMM_INFO			*comm_info;					/* channels, queues				*/
	REDUCER_INFO 		*reducer_info;				/* init stack size, sched lists	*/
	GENERAL_INFO 		*general_info;				/* misc							*/
	COPY_INFO			*copy_info;					/* garb collector status, temps	*/
	DESC_INFO			*descriptor_info;			/* offsets of important descs	*/
	CODE_INFO			*code_info;					/* special code addresses		*/

} ROOT;

/********************************************************************************
 *																				*
 *		on-chip structures														*
 *																				*
 ********************************************************************************/

typedef WORD SYSTEM_STACK[25];			/* for runtime system only	*/
typedef WORD PROCESS_STACK[16];			/* for reducers only		*/

typedef struct VRED_ON_CHIP_DATA {

	/* the first fields need to be at fixed locations!						*/
	
	WORD			stop_wsp;			/* used for increasing priority		*/
	SIGNUM			|signal_channel;	/* for signals to runtime system	*/
	ROOT			*root;				/* root of runtime system			*/
	ADDRESS			descriptors;		/* application descriptor table		*/
	
	/* the following fields do not have to be at fixed locations			*/
	
	SYSTEM_STACK	system_stack;		/* stack for runtime system			*/
	PROCESS_STACK	process_stack;		/* stack for processes				*/
	
	REDUCER_LOCALS	reducer_locals[VariableArraySize];
	
} VRED_ON_CHIP_DATA;

#define sys_stop_wsp		((WSP *)hard_stop_wsp())
#define sys_signal_address	((SIGNUM |*)hard_signal_address())

#define sys_root			((ROOT *)hard_root())
#define sys_descriptors		((ADDRESS)hard_descriptors())

/********************************************************************************
 *																				*
 *		statically allocated runtime structures									*
 *																				*
 ********************************************************************************/

typedef struct SYS
{
	/* combination of all stuff reachable by reducers */

	struct PROCESSOR_STATE	processor_state;
	struct DEBUGGER_DATA	debug;
	
	struct ROOT				root;

	struct COMM_INFO		comm_info;
	struct REDUCER_INFO		reducer_info;
	struct GENERAL_INFO 	general_info;
	struct COPY_INFO		copy_info;
	struct DESC_INFO		descriptor_info;
	struct CODE_INFO		code_info;

} SYS;

/********************************************************************************
 *																				*
 *		access macros															*
 *																				*
 ********************************************************************************/

#define table_refnum(id)			(sys.comm_info.id_table[id].graph.refnum)

#define table_graph(id)				(sys.comm_info.id_table[id].graph.graph)
#define table_graph_reserved(id)	(sys.comm_info.id_table[id].graph.reserved)
#define table_graph_requests(id)	(sys.comm_info.id_table[id].graph.requests)
#define table_graph_reducer(id)		(sys.comm_info.id_table[id].graph.pid)

#define table_discarded(id)			(sys.comm_info.id_table[id].alloc_fail.discarded)

#define	reserved_graph(id)			((VHEAP_GRAPH *)(sys.comm_info.id_table[id].graph.reserved->string_node.string));

#define table_idbuff1(id)			(sys.comm_info.id_table[id].idbuff.buff1)
#define table_idbuff2(id)			(sys.comm_info.id_table[id].idbuff.buff2)
#define table_idwait(id)			(sys.comm_info.id_table[id].idbuff.wait)

/********************************************************************************
 *																				*
 *		system parameters														*
 *																				*
 ********************************************************************************/

typedef struct NET_INFO
{
	WORD Nclass;				/* computed locally */
	WORD Nbuffs;				/* computed locally */
	WORD diameter;				/* distributed by root after computation rout info */
	WORD size;					/* initialised after floodfill */
	
} NET_INFO;

typedef struct PROCESSOR_INFO {

	/* variables, adjusted during run-time (except id) */
	
	UWORD	id;								/* constant: identifies address of data */
	
	TICKS	ticks_collecting;
	TICKS	ticks_running;
	TICKS	ticks_copying;
	
	UWORD	n_in_bytes[VLinks];
	UWORD	n_out_bytes[VLinks];
	
	WORD	error_number;					/* current local error number */
	BOOL	fatal_error_flag;				/* true if fatal error occured */
	BOOL	error_flag;						/* true if error occured */
	BOOL	warning_flag;					/* true if warning occured */
	
	ADDRESS	program_start;					/* start of current program */
	UWORD	program_size;					/* size of current program */
	
	WORD	n_procs_initialised;			/* amount of intialised procs */
	
} PROCESSOR_INFO;

typedef struct NODE_INFO
{
	/* constants, initialised during start-up */
	
	WORD				id;
	
	/* set by memory module */
	
	VRED_ON_CHIP_DATA	*int_mem_start;
	WORD				int_mem_size;
	WORD				ext_mem_size;		/* options can override test */

	/* set by netboot */
		
	CHANNEL 			*to_host;
	CHANNEL 			*from_host;
		
	WORD				nbooted[NLinks];
	WORD				neighbour[NLinks];

	/* set by router */
	
	DESTINATION_INFO	*rout_data;
		
	WORD				max_reducers;
	UWORD				scheduling_bias;

	ADDRESS				low_valid_address;
	ADDRESS				high_valid_address;

	WORD				machine_name_length;
	CHAR				machine_name[80];

} NODE_INFO;

/********************************************************************************
 *																				*
 *		options, distributed by runtime system from root						*
 *																				*
 ********************************************************************************/

typedef struct OPTIONS
{
	/* distributed by root after reading commandline options */
	
	WORD heap_size;
	WORD ext_mem_size;			/* can be used to override test (default 0 means: perform test) */
	
	WORD init_stack_size;
	WORD init_heap_size;
	WORD increase_heap_size;
	WORD init_id_table_size;
	WORD swap_time;
	
	WORD verbose;
	WORD extra_verbose;
	WORD show_garbage_collection;
	WORD check;
	WORD basic_value_only;
	
	WORD fixed_random;
	WORD line_buffering;
	WORD root_processor;
	
	/* options for packet router */
	
	WORD n_buffers;
	WORD packet_size;
	
	/* options for id buffer */
	
	WORD idbuffsize;

} OPTIONS;

#endif
